//==================================================================================
// LCMinigun
//==================================================================================
class LCMinigun extends minigun2;

//OBFSTART
var XC_CompensatorChannel LCChan;
var bool bBulletNow;

replication
{
	reliable if ( Role == ROLE_Authority )
		FixOffset;
}

////////////////////////////////
//All of the unlagged code here
simulated event KillCredit( actor Other)
{
	if ( XC_CompensatorChannel(Other) != none )
		LCChan = XC_CompensatorChannel(Other);
}
simulated function bool IsLC()
{
	return (LCChan != none) && LCChan.bUseLC && (LCChan.Owner == Owner);
}
function SetHand (float hand)
{
	Super.SetHand(hand);
	FixOffset(FireOffset.Y);
}
simulated function FixOffset (float Y)
{
	FireOffset.Y=Y;
}


//Serverside tracer
function TraceFire( float Accuracy )
{
	local vector HitLocation, HitNormal, StartTrace, EndTrace, X,Y,Z, AimDir;
	local actor Other, aActor;

	if ( !IsLC() )
	{
		Super.TraceFire( Accuracy);
		return;
	}
	if ( Owner == none ) return;

	Owner.MakeNoise(Pawn(Owner).SoundDampening);
	GetAxes(Pawn(owner).ViewRotation,X,Y,Z);
	StartTrace = Owner.Location + CalcDrawOffset() + FireOffset.Y * Y + FireOffset.Z * Z; 
	AdjustedAim = pawn(owner).AdjustAim(1000000, StartTrace, 2.75*AimError, False, False);	
	EndTrace = StartTrace + Accuracy * (FRand() - 0.5 )* Y * 1000 + Accuracy * (FRand() - 0.5 ) * Z * 1000;
	AimDir = vector(AdjustedAim);
	EndTrace += (10000 * AimDir);
	LCChan.LCActor.ffUnlagPositions( LCChan.LCComp);
	ForEach Owner.TraceActors( class'Actor', aActor, HitLocation, HitNormal, EndTrace, StartTrace)
	{
		if ( Class'LCStatics'.static.TraceStopper( aActor) )
		{
			Other = aActor;
			break;
		}
		if ( !aActor.bProjTarget && !aActor.bBlockActors )
			continue;
		if ( aActor.IsA('Pawn') && (Pawn(aActor).PlayerReplicationInfo != none || !Pawn(aActor).AdjustHitLocation(HitLocation, EndTrace - StartTrace)) )
			continue; //Pawns with PRI are ignored, this skips players and prevents double hitbox during lag compensation

		Other = aActor;
		break;
	}
//	Other = Pawn(Owner).TraceShot(HitLocation,HitNormal,EndTrace,StartTrace);

	Count++;
	if ( Count == 4 )
	{
		Count = 0;
		if ( VSize(HitLocation - StartTrace) > 250 )
			Spawn(class'LCMTracer',Owner,, StartTrace + 96 * AimDir,rotator(EndTrace - StartTrace));
	}
	ProcessTraceHit2(Other, HitLocation, HitNormal, vector(AdjustedAim),Y,Z);
	LCChan.LCActor.ffRevertPositions();
}

function ProcessTraceHit2(Actor Other, Vector HitLocation, Vector HitNormal, Vector X, Vector Y, Vector Z)
{
	local int rndDam;

	if (Other == Level) 
		Spawn(class'LCLightWallHitEffect',Owner,, HitLocation+HitNormal, Rotator(HitNormal));
	else if ( (Other!=self) && (Other!=Owner) && (Other != None) ) 
	{
		if ( Other.IsA('XC_LagCompensator') ) //I just hit a lag compensator, apply damage to Pawn
		{
			HitLocation -= Other.Location;
			Other = XC_LagCompensator(Other).ffOwner;
			HitLocation += Other.Location;
			Other.PlaySound(Sound 'ChunkHit',, 4.0,,100);
		}
		else if ( !Other.bIsPawn && !Other.IsA('Carcass') )
			spawn(class'UT_SpriteSmokePuff',,,HitLocation+HitNormal*9); 
		else
			Other.PlaySound(Sound 'ChunkHit',, 4.0,,100);

		if ( Other.IsA('Bot') && (FRand() < 0.2) )
			Pawn(Other).WarnTarget(Pawn(Owner), 500, X);
		rndDam = 9 + Rand(6);
		if ( FRand() < 0.2 )
			X *= 2.5;
		Other.TakeDamage(rndDam, Pawn(Owner), HitLocation, rndDam*500.0*X, MyDamageType);
	}
}


simulated state ClientFiring
{
Begin:
	ShotAccuracy = 0.2;
	Sleep(0.13);
	bBulletNow = true;
	Goto('Begin');
}

simulated state ClientAltFiring
{
Begin:
	Sleep(0.13);
	bBulletNow = true;
	if ( AnimSequence == 'Shoot2' )
		Goto('FastShoot');
	Goto('Begin');
FastShoot:
	ShotAccuracy = 0.75;
	Sleep(0.08);
	bBulletNow = true;
	Goto('FastShoot');
}

//Doing this because ACE has bugs, problem is, tick happens 1 frame after bBulletNow is set!!!
simulated event Tick( float DeltaTime)
{
	Super.Tick(DeltaTime);
	if ( bBulletNow )
	{
		SimGenerateBullet();
		bBulletNow = false;
	}
}

simulated function SimGenerateBullet()
{
	if ( !IsLC() )
		return;
	SimTraceFire( ShotAccuracy);
}

simulated function SimTraceFire( float Accuracy )
{
	local vector HitLocation, HitNormal, StartTrace, EndTrace, X,Y,Z;
	local actor Other, aActor;

	if ( Owner == none )
		return;
	GetAxes( class'LCStatics'.static.PlayerRot( Pawn(Owner)), X,Y,Z);

	StartTrace = Owner.Location + CalcDrawOffset() + FireOffset.Y * Y + FireOffset.Z * Z;  //CALCDRAWOFFSET MIGHT SCREW UP THINGS
	EndTrace = StartTrace + Accuracy * (FRand() - 0.5 )* Y * 1000 + Accuracy * (FRand() - 0.5 ) * Z * 1000;
	EndTrace += (10000 * X); 
	ForEach Owner.TraceActors( class'Actor', aActor, HitLocation, HitNormal, EndTrace, StartTrace)
	{
		if ( Class'LCStatics'.static.TraceStopper( aActor) )
		{	Other = aActor;
			break;
		}
		if ( !aActor.bProjTarget && !aActor.bBlockActors )
			continue;
		if ( aActor.IsA('Pawn') && !Pawn(aActor).AdjustHitLocation(HitLocation, EndTrace - StartTrace) )
			continue;
		Other = aActor;
		break;
	}

	Count++;
	if ( Count == 4 )
	{
		Count = 0;
		if ( VSize(HitLocation - StartTrace) > 250 )
			Spawn(class'MTracer',,, StartTrace + 96 * X,rotator(EndTrace - StartTrace));
	}
	if (Other == Level) 
		Spawn(class'UT_LightWallHitEffect',Owner,, HitLocation+HitNormal, Rotator(HitNormal));
}


state NormalFire
{
	function Tick( float DeltaTime )
	{
		if (Owner==None) 
		{
			AmbientSound = None;
			return;
		}

		if ( Pawn(owner).PlayerReplicationInfo.bFeigningDeath )
			GotoState('FinishFire');

	}
}

state AltFiring
{
	function Tick( float DeltaTime )
	{
		if (Owner==None) 
		{
			AmbientSound = None;
			GotoState('Pickup');
			return;
		}			

		if ( bFiredShot && ((pawn(Owner).bAltFire==0) || bOutOfAmmo) ) 
			GoToState('FinishFire');

		if ( Pawn(owner).PlayerReplicationInfo.bFeigningDeath )
			GotoState('FinishFire');
	}
}


defaultproperties
{
}
