//=============================================================================
// XC_LagCompensation.
// Max supported ping = 650
// Made by Higor
//=============================================================================
class XC_LagCompensation expands Mutator;


var XC_LagCompensator ffCompList;
var XC_PlayerPos ffPosPool;
var XC_ComboTracker ffComboPool, ffUsedCombos;
var int iCombo; //Combo tracker count
var float ffMaxLatency;
var string Pkg;
var int ffCurId;

var XC_PlayerPos ffPosTemp;


event Tick( private float ffDelta)
{
	ffPosTemp = none; //Clear
}


function ffCollidePlayers( private bool ffEnable)
{
	local private XC_LagCompensator ffTmp;
	
	if ( !ffEnable )
	{	For ( ffTmp=ffCompList ; ffTmp!=none ; ffTmp=ffTmp.ffCompNext )
		{
			ffTmp.ffGetCollision();
			ffTmp.ffNoCollision();
		}
	}
	else
		For ( ffTmp=ffCompList ; ffTmp!=none ; ffTmp=ffTmp.ffCompNext )
			ffTmp.ffSetCollision();
}

function bool ffInsertNewPlayer( private pawn ffOther)
{
	local private XC_LagCompensator ffTmp;
	
	if ( ffOther == none )
		return false;
	ffTmp = Spawn( class'XC_LagCompensator', ffOther);
	ffTmp.ffOwner = ffOther;
	ffTmp.ffCompNext = ffCompList;
	ffCompList = ffTmp;
	ffTmp.ffMaster = self;

	//Register individual ZP controller channel on this playerpawn
	if ( (PlayerPawn(ffOther) != none) && (NetConnection(PlayerPawn(ffOther).Player) != none) )
		Spawn(class'XC_CompensatorChannel').AddPlayer( PlayerPawn(ffOther), self);
}

function XC_LagCompensator ffFindCompFor( private pawn ffOther)
{
	local private XC_LagCompensator ffTmp;
	
	For ( ffTmp=ffCompList ; ffTmp!=none ; ffTmp=ffTmp.ffCompNext )
		if ( ffTmp.ffOwner == ffOther )
			return ffTmp;
	return none;
}

function XC_LagCompensator ffGetLC( private pawn ffOther)
{
	local private XC_LagCompensator ffTmp;

	For ( ffTmp=ffCompList ; ffTmp!=none ; ffTmp=ffTmp.ffCompNext )
		if ( ffTmp.ffOwner == ffOther )
			return ffTmp;
	return none;
}

function XC_PlayerPos ffRequestPos( private XC_LagCompensator ffOther)
{
	local private XC_PlayerPos ffTmp;
	if ( ffPosPool != none )
	{
		ffTmp = ffPosPool;
		ffPosPool = ffPosPool.ffNextPos;
		ffTmp.ffNextPos = none;
	}
	else
	{
		ffCurId++;
		ffTmp = Spawn( class'XC_PlayerPos');
	}
	ffTmp.ffOwner = ffOther;
	return ffTmp;
}

function ffUnlagShockBalls( private XC_LagCompensator ffOther)
{
	local private float ffPing;
	local private ShockProj ffTmp;
	local private XC_ComboTracker ffPos;

	ffPing = (float(ffOther.ffLastPing) / 1000.0) * Level.TimeDilation;
	ForEach AllActors (class'ShockProj', ffTmp)
	{
		if ( ffComboPool == none )
		{
			ffComboPool = Spawn(class'XC_ComboTracker');
			iCombo++;
		}
		ffPos = ffComboPool;
		if ( ffPos.ffTrack( ffTmp, ffPing) )
		{
			ffComboPool = ffPos.nextTracker;
			ffPos.nextTracker = ffUsedCombos;
			ffUsedCombos = ffPos;
		}
	}
}

function ffResetShockBalls()
{
	local private ShockProj ffTmp;
	local private XC_ComboTracker ffPos;
	local private int nani;

	if ( ffUsedCombos == none )
		return;

	ffPos = ffUsedCombos;
	While ( ffPos != none )
	{
		ffPos.ffReset();
		ffPos = ffPos.nextTracker;
		nani++;
	}

	if ( ffComboPool == none )
	{
		ffComboPool = ffUsedCombos;
		ffComboPool = none;
	}
	else if ( iCombo - nani < nani ) //OPTIMIZATION: Active trackers are many, loop on passive trackers, add used combos to end of pool
	{
		ffPos = ffComboPool;
		while ( ffPos.nextTracker != none )
			ffPos = ffPos.nextTracker;
		ffPos.nextTracker = ffUsedCombos;
		ffUsedCombos = none;
	}
	else //OPTIMIZATION: Passive trackers are many, loop on active trackers, append pool after it, then move all to pool
	{
		ffPos = ffUsedCombos;
		while ( ffPos.nextTracker != none )
			ffPos = ffPos.nextTracker;
		ffPos.nextTracker = ffComboPool;
		ffComboPool = ffUsedCombos;
		ffUsedCombos = none;
	}
}

function bool ffUnlagPositions( private XC_LagCompensator ffOther)
{
	local private float ffPing, ffDelta;
	local private XC_PlayerPos ffTmp;
	local private vector ffPos;
	local private XC_LagCompensator nani2;

	if ( ffOther == none )
		return false;

	ffPing = float(ffOther.ffLastPing) / 1000.0;
	nani2 = ffCompList;
	while ( nani2 != none )
	{
		ffTmp = nani2.ffFindNewer( ffPing);
		if ( ffTmp != none )
			break;
		nani2 = nani2.ffCompNext;
	}

	For ( nani2=none ; ffTmp != none ; ffTmp = ffTmp.ffContiguous )
	{
		if ( ffTmp.ffOwner == ffOther || ffTmp.ffGhost )
			continue;
		if ( (ffTmp.ffNextPos == none) || ffTmp.ffTeleported )
			ffPos = ffTmp.Location;
		else
		{
			if ( ffDelta == 0 ) //ffDelta needs to be prepared only once
			{
				ffDelta = (ffTmp.ffSTimeStamp - ffTmp.ffNextPos.ffSTimeStamp) / Level.TimeDilation;
				ffPing -= (Level.TimeSeconds - ffTmp.ffSTimeStamp) / Level.TimeDilation;
				ffDelta = ffPing / ffDelta;
			}
			ffPos = ffTmp.Location + (ffTmp.ffNextPos.Location - ffTmp.Location) * ffDelta;
		}
		if ( ffTmp.ffDuck )
		{
			ffPos.Z -= ffTmp.ffOwner.ffOwner.CollisionHeight * 0.7;
			ffTmp.ffOwner.SetCollisionSize( ffTmp.ffOwner.ffOwner.CollisionRadius, ffTmp.ffOwner.ffOwner.CollisionHeight * 0.7);
		}
		else
			ffTmp.ffOwner.SetCollisionSize( ffTmp.ffOwner.ffOwner.CollisionRadius, ffTmp.ffOwner.ffOwner.CollisionHeight);
		ffTmp.ffOwner.SetLocation( ffPos);
		ffTmp.ffOwner.SetCollision( true, false, false);
		ffTmp.ffOwner.bProjTarget = true;
	}
	return true;
}

function ffRevertPositions()
{
	local private XC_LagCompensator ffTmp;
	
	For ( ffTmp = ffCompList ; ffTmp!=none ; ffTmp=ffTmp.ffCompNext )
		if ( ffTmp.bProjTarget )
		{
			ffTmp.bProjTarget = false;
			ffTmp.SetCollision(false,false,false);
		}
}

function float ffMaxLag()
{
	return ffMaxLatency * 0.001 * Level.TimeDilation; //Scale up if game is faster, we'll be operating using just ffDelta on other calculation phases
}
//OBFEND

//Fast validation checks
function bool FastValidate( XC_CompensatorChannel LCChan, Actor Other, int HashID, Weapon aWeap, int CmpRot, vector PLoc, vector Start, vector End, out byte Imprecise, float Accuracy, int Flags)
{
	local vector aVec;
	local pawn P;
	local rotator FixedRot;
	local float fDist;

	//One or the other
	if ( Other != none && HashID != -1 )
		return false;
	P = Pawn(LCChan.Owner);
	if ( P.Health < 0 )
		return false;
	if ( (P == none) || (P.Weapon != aWeap) )
	{
		Log("WRONG WEAPON", 'LagCompensator');
		return false;
	}
	//No Seed with accuracy = hack?
	if ( Accuracy != 0 )
	{
		if ( (Flags >>> 16) == 0 )
			return false;
		if ( string(Accuracy) != aWeap.GetPropertyText("ffAimError") ) //Weapon aim error mismatch
		{
			Log("Aim error mismatch:"@Accuracy@"vs"@aWeap.GetPropertyText("ffAimError"),'LagCompensator');
			return false;
		}
	}
	if ( aWeap.IsInState('DownWeapon') )
	{
		Log("DOWNWEAPON State", 'LagCompensator');
		return false;
	}
	if ( Accuracy > 0 )
	{
		fDist = 10000;
		if ( (Flags & 4) > 0 )			fDist += 10000;
		if ( (Flags & 8) > 0 )			fDist += 20000;
		aVec = ValidateEnd( Accuracy, fDist ,Flags >>> 16, class'LCStatics'.static.DecompressRotator(CmpRot));
	}
	else
		aVec = Vector(class'LCStatics'.static.DecompressRotator(CmpRot));
	if ( (VSize(End-Start) > 30) && (VSize( aVec - Normal(End-Start)) > 0.05) )
	{
		if ( (Imprecise > 0) || (VSize( aVec - Normal(End-Start)) > 0.20) )
		{
			Log( "DIRECTION DIFF IS :"$ VSize( aVec - Normal(End-Start)) , 'LagCompensator');
			return false;
		}
		else Imprecise++;	
	}
	if ( Accuracy > 0 )
		aVec = Vector(class'LCStatics'.static.DecompressRotator(CmpRot));
	if ( (VSize( aVec - Vector(P.ViewRotation)) > 0.07) && (VSize( aVec - Vector(LCChan.OldView)) > 0.07) )
	{
		FixedRot = rotator(vector(P.ViewRotation) + vector(LCChan.OldView));
		if ( VSize( aVec - Vector(FixedRot)) < 0.07 )
		{}
		else if ( (Imprecise > 0) || (VSize( aVec - Vector(FixedRot)) > 0.22) )
		{
			Log( "VROTATION DIFFERENCE IS: "$ VSize( aVec - Vector(P.ViewRotation)) , 'LagCompensator');
			return false;
		}
		else Imprecise++;
	}
	aVec.X = VSize(P.Location - P.OldLocation); //Tick loc difference, helps low Rate servers
	if ( P.Base != none )
		aVec.X += VSize(P.Base.Velocity) * 0.10;
	if ( P.Physics == PHYS_Walking )
		aVec.X += 20 + VSize(P.Velocity) * 0.05;
	if ( VSize(WeaponStartTrace(aWeap) - Start) > (50 + aVec.X + VSize(P.Velocity) * 0.20) )
	{
		if ( (Imprecise > 0) || (VSize(WeaponStartTrace(aWeap) - Start) > (75 + aVec.X * 1.1 + VSize(P.Velocity) * 0.32)) )
		{
			Log( "VECTOR START DIFF = "$VSize(WeaponStartTrace(aWeap) - Start) , 'LagCompensator');
			return false;
		}
		else Imprecise++;
	}
	if ( VSize( PLoc - P.Location) > (45 + aVec.X + VSize(P.Velocity) * 0.15) )
	{
		if ( (Imprecise > 0) || (VSize( PLoc - P.Location) > (60 + aVec.X * 1.1 + VSize(P.Velocity) * 0.28)) )
		{
			Log( "LOCATION DIFF = "$VSize( PLoc - P.Location) , 'LagCompensator');
			return false;
		}
		else Imprecise++;
	}
	return true;
}

function vector ValidateEnd( float Accuracy, float MaxDist, int Seed, rotator Dir)
{
	local vector X,Y,Z;

	GetAxes( Dir, X,Y,Z);
	return Normal(X * MaxDist + class'LCStatics'.static.StaticAimError( Y, Z, Accuracy, Seed));
}


function vector WeaponStartTrace( Weapon W)
{
	local vector X,Y,Z;
	GetAxes( Pawn(W.Owner).ViewRotation, X, Y, Z);
	return W.Owner.Location + W.CalcDrawOffset() + W.FireOffset.Y * Y + W.FireOffset.Z * Z; 
}


defaultproperties
{
    bGameRelevant=True
	ffMaxLatency=650
	RemoteRole=ROLE_None
}
